/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sonar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.sonar.plugins.cobertura;

import ch.hortis.sonar.service.MeasureKey;
import org.apache.commons.lang.StringUtils;
import org.apache.maven.plugin.logging.Log;
import org.hibernate.Hibernate;
import org.sonar.commons.measures.MeasureParameter;
import org.sonar.plugins.api.maven.*;
import org.sonar.plugins.api.maven.model.MavenPom;
import org.sonar.plugins.api.maven.xml.XmlParserException;
import org.sonar.plugins.api.maven.xml.XmlReportParser;
import org.sonar.plugins.api.metrics.CoreMetrics;
import org.sonar.plugins.core.unittests.CodeCoverageUtils;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.io.File;
import java.sql.Blob;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

public class CoberturaMavenCollector extends AbstractJavaMavenCollector implements CollectsCodeCoverage {
  public static final String PROP_KEY_REUSE_XML = "sonar.cobertura.useBuildXmlResults";

  public CoberturaMavenCollector(JavaMeasuresRecorder recorder, Log log) {
    super(recorder, log);
  }

  public Class<? extends MavenPluginHandler> dependsOnMavenPlugin(MavenPom pom) {
    if (pom.isSonarLightMode()) {
      return null;
    }
    return CoberturaMavenPluginHandler.class;
  }

  public boolean shouldStopOnFailure() {
    return true;
  }

  protected boolean shouldCollectIfNoSources() {
    return false;
  }

  public void collect(MavenPom pom) {
    String reuse = pom.getConfiguration().getString(PROP_KEY_REUSE_XML);
    if ( !pom.isSonarLightMode() || reuse!=null) {
      File reportXmlFile = MavenCollectorUtils.findFileFromBuildDirectory(pom, "site/cobertura/coverage.xml");
      collect(reportXmlFile);
    }
  }

  protected void collect(File reportXmlFile) {
    try {
      if (reportXmlFile != null) {
        XmlReportParser parser = new XmlReportParser();
        parser.parse(reportXmlFile);

        collectProjectMeasures(parser);
        collectPackageMeasures(parser);
        collectClassMeasures(parser);
      }

    } catch (ParseException e) {
      throw new XmlParserException(e);
    }
  }

  private void collectProjectMeasures(XmlReportParser parser) throws ParseException {
    double lineRate = MavenCollectorUtils.parseNumber(parser.executeXPath("/coverage/@line-rate"), Locale.ENGLISH);
    getMeasuresRecorder().createProjectMeasure(new MeasureKey(CoreMetrics.CODE_COVERAGE), convertPercentage(lineRate));
  }

  private void collectPackageMeasures(XmlReportParser
    parser) throws ParseException {
    NodeList packages = parser.executeXPathNodeList("//packages/package");
    for (int i = 0; i < packages.getLength(); i++) {
      Element elt = (Element) packages.item(i);
      getMeasuresRecorder().createPackageMeasure(elt.getAttribute("name"), new MeasureKey(CoreMetrics.CODE_COVERAGE),
        convertPercentage(MavenCollectorUtils.parseNumber(elt.getAttribute("line-rate"), Locale.ENGLISH)));
    }
  }

  private void collectClassMeasures(XmlReportParser parser) throws ParseException {
    NodeList classes = parser.executeXPathNodeList("//classes/class");
    for (int i = 0; i < classes.getLength(); i++) {
      Element elt = (Element) classes.item(i);
      recordClassMeasure(parser, elt, "line-rate");
    }
  }

  private void recordClassMeasure(XmlReportParser parser, Element elt, String valueAttributeName) throws ParseException {
    String name = elt.getAttribute("name");
    String classname = name;
    String packagename = null;
    if (name.indexOf(".") >= 0) {
      classname = StringUtils.substringAfterLast(name, ".");
      packagename = StringUtils.substringBeforeLast(name, ".");
    }
    getMeasuresRecorder().createClassMeasure(packagename, classname, false, new MeasureKey(CoreMetrics.CODE_COVERAGE),
      convertPercentage(MavenCollectorUtils.parseNumber(elt.getAttribute(valueAttributeName), Locale.ENGLISH)), getParameters(parser, elt));
  }

  private MeasureParameter[] getParameters(XmlReportParser parser, Element elt) throws ParseException {
    List<MeasureParameter> parameters = new ArrayList<MeasureParameter>();
    NodeList lines = parser.executeXPathNodeList(elt, "methods/method/lines/line");
    StringBuilder levelsByLine = new StringBuilder();
    for (int i = 0; i < lines.getLength(); i++) {
      Element line = (Element) lines.item(i);
      CodeCoverageUtils.appendParameter(levelsByLine, line.getAttribute("number"), line.getAttribute("hits"));
    }
    Blob lobValue = Hibernate.createBlob(levelsByLine.toString().getBytes());
    parameters.add(new MeasureParameter("lines", lobValue, null));
    return parameters.toArray(new MeasureParameter[0]);
  }

  private double convertPercentage(Number percentage) {
    return MavenCollectorUtils.scaleValue(percentage.doubleValue() * 100.0);
  }

  public void configure(CollectsUnitTests unitTestsCollector) {
    // nothing to do. Cobertura does not change location of surefire reports
  }
}
